# Generated by Django 2.2.11 on 2020-03-11 13:54

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("lti_toolbox", "0001_initial"),
        ("forum", "0011_auto_20190627_2132"),
    ]

    operations = [
        migrations.AddField(
            model_name="forum",
            name="lti_consumer",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to="lti_toolbox.LTIConsumer",
            ),
        ),
        migrations.AddField(
            model_name="forum",
            name="lti_context",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.10.1 on 2016-09-03 02:50
from __future__ import unicode_literals

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("forum_conversation", "0006_post_enable_signature"),
    ]

    operations = [
        migrations.AddField(
            model_name="topic",
            name="first_post",
            field=models.ForeignKey(
                blank=True,
                editable=False,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="+",
                to="forum_conversation.Post",
                verbose_name="Last post",
            ),
        ),
        migrations.AddField(
            model_name="topic",
            name="last_post",
            field=models.ForeignKey(
                blank=True,
                editable=False,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="+",
                to="forum_conversation.Post",
                verbose_name="Last post",
            ),
        ),
    ]



"""This application overrides django-machina's forum_permission app."""

# pylint: disable=invalid-name
default_app_config = (
    "ashley.machina_extensions.forum_permission.apps.ForumPermissionAppConfig"
)



"""Ashley forum_permission application"""
from machina.apps.forum_permission.apps import (
    ForumPermissionAppConfig as BaseForumPermissionAppConfig,
)


class ForumPermissionAppConfig(BaseForumPermissionAppConfig):
    """Configuration class for the forum_permission app."""

    name = "ashley.machina_extensions.forum_permission"



"""
This module defines URL patterns associated with the django-machina``forum``
application.
"""

from django.urls import path
from machina.apps.forum.urls import (
    ForumURLPatternsFactory as BaseForumURLPatternsFactory,
)
from machina.core.loading import get_class


class ForumURLPatternsFactory(BaseForumURLPatternsFactory):
    """Allows to generate the URL patterns of the ``forum`` application."""

    rename_view = get_class("forum.views", "ForumRenameView")

    def get_urlpatterns(self):
        """Returns the URL patterns managed by the considered factory / application."""

        return super().get_urlpatterns() + [
            path("admin/rename/<int:pk>/", self.rename_view.as_view(), name="rename"),
        ]


urlpatterns_factory = ForumURLPatternsFactory()



# Generated by Django 3.1.7 on 2021-03-02 16:14

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("forum_conversation", "0013_auto_20201220_1745"),
    ]

    operations = [
        migrations.AlterField(
            model_name="topic",
            name="posts_count",
            field=models.PositiveIntegerField(
                blank=True,
                db_index=True,
                default=0,
                editable=False,
                verbose_name="Posts count",
            ),
        ),
        migrations.AlterField(
            model_name="topic",
            name="views_count",
            field=models.PositiveIntegerField(
                blank=True,
                db_index=True,
                default=0,
                editable=False,
                verbose_name="Views count",
            ),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("forum_conversation", "0001_initial"),
    ]

    operations = [
        migrations.AddField(
            model_name="post",
            name="anonymous_key",
            field=models.CharField(
                max_length=100,
                null=True,
                verbose_name="Anonymous user forum key",
                blank=True,
            ),
        ),
    ]



"""
    Forum search views
    ==================

    This module defines views provided by the ``forum_search`` application.

"""
from haystack import views

from ashley.context_mixins import get_current_lti_session


class FacetedSearchView(views.FacetedSearchView):
    """View to show search results"""

    template = "forum_search/search.html"

    def build_form(self, form_kwargs=None):
        form = super().build_form(
            form_kwargs={
                "user": self.request.user,
                "lti_context": get_current_lti_session(self.request),
            }
        )
        return form



"""Views of the ashley api django application."""
import logging

from django.contrib.auth import get_user_model
from django.core.exceptions import PermissionDenied
from machina.apps.forum_permission.viewmixins import (
    PermissionRequiredMixin as BasePermissionRequiredMixin,
)
from machina.core.db.models import get_model
from machina.core.loading import get_class
from rest_framework import mixins, status, viewsets
from rest_framework.permissions import IsAuthenticated
from rest_framework.response import Response

from ashley.context_mixins import get_current_lti_session
from ashley.defaults import _FORUM_ROLE_INSTRUCTOR, _FORUM_ROLE_MODERATOR
from ashley.permissions import ManageModeratorPermission

from .serializers import UploadImageSerializer, UserSerializer

User = get_user_model()
LTIContext = get_model("ashley", "LTIContext")
Forum = get_model("forum", "Forum")
UploadImage = get_model("ashley", "UploadImage")
PermissionRequiredMixin: BasePermissionRequiredMixin = get_class(
    "forum_permission.viewmixins", "PermissionRequiredMixin"
)
logger = logging.getLogger(__name__)


class UserApiView(
    mixins.ListModelMixin, mixins.UpdateModelMixin, viewsets.GenericViewSet
):  # pylint: disable=too-many-ancestors
    """API requests to list moderators for current LTIContext"""

    # Add permission to check if user has the right permission and can manage moderators
    permission_classes = [ManageModeratorPermission]
    serializer_class = UserSerializer
    ordering = "public_username"

    def get_queryset(self):
        context = get_current_lti_session(self.request)
        role = self.request.GET.get("role", None)
        # Select all active users who are part of the current LTIContext
        query = User.objects.filter(
            is_active=True,
            groups__name=context.base_group_name,
        )
        # If a role is specified and start with !, we exclude its members from the results
        if role is not None and role.startswith("!"):
            query = query.exclude(groups__name=context.get_group_role_name(role[1:]))
        # If a role is specified, we select only users having this role inside the LTIContext
        elif role:
            query = query.filter(groups__name=context.get_group_role_name(role))
        # When the role moderator (or !moderator) is specified, we want to exclude the
        # users having the instructor role from the results.
        if role in [_FORUM_ROLE_MODERATOR, f"!{_FORUM_ROLE_MODERATOR}"]:
            query = query.exclude(
                groups__name=context.get_group_role_name(_FORUM_ROLE_INSTRUCTOR)
            )
        return query.order_by(self.ordering)

    # pylint: disable=W0221
    def update(self, request, pk=None):
        """Update method is used to promote or revoke moderator role"""
        user = User.objects.get(pk=pk)
        lti_context = get_current_lti_session(self.request)
        roles = self.request.data.get("roles", [])
        # Load current groups
        user_groups = lti_context.get_user_roles(user)
        group_moderator = lti_context.get_role_group(_FORUM_ROLE_MODERATOR)

        # Asked to revoke moderator,
        if _FORUM_ROLE_MODERATOR not in roles and _FORUM_ROLE_MODERATOR in user_groups:
            user.groups.remove(group_moderator)
            user.save()
        # Asked to become moderator, check user is not yet moderator and make sure user
        # belongs to this context by checking he has groups from this context.
        elif (
            _FORUM_ROLE_MODERATOR in roles
            and _FORUM_ROLE_MODERATOR not in user_groups
            and len(user_groups) > 0
        ):
            user.groups.add(group_moderator)
            user.save()
        else:
            logger.debug(
                "Update moderator forbidden - Role asked %s - user groups %s",
                roles,
                user_groups,
            )
            raise PermissionDenied()

        return Response({"success": True}, status=status.HTTP_200_OK)


# pylint: disable=too-many-ancestors
class ImageUploadApiView(mixins.CreateModelMixin, viewsets.GenericViewSet):
    """Viewset for the API of the Image object."""

    queryset = UploadImage.objects.all()
    serializer_class = UploadImageSerializer
    permission_classes = [IsAuthenticated]

    def perform_create(self, serializer):
        """Preset poster field with current user"""
        serializer.save(poster=self.request.user)



"""Declare the models related to the forum_conversation app ."""
from django.contrib.auth import get_user_model
from django.db import models  # pylint: disable=all
from django.utils.translation import gettext_lazy as _
from machina.apps.forum_conversation.abstract_models import (
    AbstractTopic as MachinaAbstractTopic,
)
from machina.core.db.models import model_factory
from machina.core.loading import get_class

get_forum_member_display_name = get_class(
    "forum_member.shortcuts", "get_forum_member_display_name"
)
User = get_user_model()


class AbstractTopic(MachinaAbstractTopic):
    # The number of posts included in this topic (only those that are approved).
    # Add Index for sorting
    posts_count = models.PositiveIntegerField(
        editable=False,
        blank=True,
        default=0,
        verbose_name=_("Posts count"),
        db_index=True,
    )

    # The number of time the topic has been viewed.
    # Add Index for sorting
    views_count = models.PositiveIntegerField(
        editable=False,
        blank=True,
        default=0,
        verbose_name=_("Views count"),
        db_index=True,
    )

    class Meta(MachinaAbstractTopic.Meta):
        abstract = True

    def get_active_users(self, user):
        # collect active users for this topic and exclude current user
        active_post_users = (
            User.objects.filter(is_active=True, posts__topic=self, posts__approved=True)
            .exclude(pk=user.pk)
            .distinct()
        )

        list_active_users = [
            {
                "name": get_forum_member_display_name(user),
                "user": user.pk,
            }
            for user in active_post_users
        ]

        return sorted(list_active_users, key=lambda i: i["name"])


Topic = model_factory(AbstractTopic)

from machina.apps.forum_conversation.models import *  # noqa



# Generated by Django 2.2.4 on 2019-08-21 08:47

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("forum_permission", "0004_auto_20190319_2240"),
    ]

    operations = [
        migrations.AddField(
            model_name="userforumpermission",
            name="authenticated_user",
            field=models.BooleanField(
                db_index=True, default=False, verbose_name="Target any logged in user"
            ),
        ),
    ]



"""Custom permission classes for the Ashley project."""
from django.core.exceptions import PermissionDenied
from machina.apps.forum_permission.viewmixins import (
    PermissionRequiredMixin as BasePermissionRequiredMixin,
)
from machina.core.db.models import get_model
from machina.core.loading import get_class

from ashley.context_mixins import get_current_lti_session

from . import SESSION_LTI_CONTEXT_ID

PermissionRequiredMixin: BasePermissionRequiredMixin = get_class(
    "forum_permission.viewmixins", "PermissionRequiredMixin"
)
Forum = get_model("forum", "Forum")


class ManageModeratorPermission(PermissionRequiredMixin):
    """
    Permission based on LTIContext Session.
    Class used by API and view on managing moderators for current session LTIContext.
    """

    @classmethod
    def _get_forum(cls, request):
        """LTIContext can have multiple forum, we target the first one"""
        try:
            if request.user.is_authenticated and request.session.get(
                SESSION_LTI_CONTEXT_ID
            ):
                context = get_current_lti_session(request)
                return Forum.objects.filter(lti_contexts=context).first()
            return None
        except Exception as no_forum:
            raise PermissionDenied() from no_forum

    @classmethod
    def _can(cls, request):
        """Check user has permission `can_manage_moderator` for first forum in this LTIContext"""
        return request.forum_permission_handler.can_manage_moderator(
            ManageModeratorPermission._get_forum(request), request.user
        )

    # pylint: disable=unused-argument
    def has_permission(self, request, view):
        """Check user has permission `can_manage_moderator` for first forum in this LTIContext"""
        return self._can(request)

    # pylint: disable=unused-argument
    def perform_permissions_check(self, user, obj, perms):
        """Check user has permission `can_manage_moderator` for first forum in this LTIContext"""
        return self._can(self.request)

    def has_object_permission(self, request, view, obj):
        """Check user has permission `can_manage_moderator` for first forum in this LTIContext"""
        return self._can(request)



"""Ashley is a django application that customizes and adds features to the django-machina forum."""


# pylint: disable=invalid-name
import os

default_app_config = "ashley.apps.AshleyConfig"

# Main Ashley template directory.
ASHLEY_MAIN_TEMPLATE_DIR = os.path.join(
    os.path.dirname(os.path.abspath(__file__)),
    "templates",
)

# Name of the session entry containing the current LTI context ID
SESSION_LTI_CONTEXT_ID = "ltictx"



"""Serializer for ashley api."""
from django.contrib.auth import get_user_model
from machina.core.db.models import get_model
from rest_framework import serializers

from ashley.context_mixins import get_current_lti_session

User = get_user_model()
UploadImage = get_model("ashley", "UploadImage")


class UserSerializer(serializers.ModelSerializer):
    """Serializer for User model."""

    roles = serializers.SerializerMethodField()

    class Meta:
        model = User
        fields = ["id", "public_username", "roles"]
        read_only_fields = ["id", "public_username", "roles"]

    def get_roles(self, obj):
        """Describes the role of the user"""
        if obj.is_active:
            lti_context = get_current_lti_session(self.context.get("request"))
            return lti_context.get_user_roles(obj)

        return None


class UploadImageSerializer(serializers.ModelSerializer):
    """Serializer for Image Upload model."""

    class Meta:
        model = UploadImage
        fields = (
            "id",
            "file",
            "forum",
            "poster",
        )
        extra_kwargs = {"poster": {"required": False}}



"""Authentication backend for Ashley"""
import hashlib
import logging
from typing import List

from django.contrib.auth import get_user_model
from django.core.exceptions import PermissionDenied
from lti_toolbox.backend import LTIBackend as ToolboxLTIBackend
from lti_toolbox.lti import LTI

logger = logging.getLogger(__name__)


class LTIBackend(ToolboxLTIBackend):
    """
    Authentication backend using LTI launch request to authenticate a user.
    If a user does not exist in database, it creates it.
    The username is generated with the remote (consumer) user id and with
    the consumer identifier, to be able to authenticate users from multiple
    consumer sites without any collision.
    """

    def authenticate(self, request, username=None, password=None, **kwargs):
        lti_request: LTI = kwargs.get("lti_request")
        if not lti_request:
            return None

        if not lti_request.is_valid:
            raise PermissionDenied()

        lti_consumer = lti_request.get_consumer()
        if not lti_consumer:
            raise PermissionDenied

        remote_user_id = self._get_remote_user_id(lti_request)

        # Try to get optional values from LTI request
        email = lti_request.get_param("lis_person_contact_email_primary", "")
        public_username = self._get_public_username(lti_request)

        # Handle special case for OpenedX studio :
        # When OpenedX studio initiates a LTI request to render ashley in a
        # xblock, it always sends "student" as the user_id. This is a problem,
        # because all openedX studio users of the same LTI consumer would
        # share the same user account on ashley. And they would have an
        # instructor access to every forums related to their LTI consumer.
        # To resolve this problem, we generate a user id based on the
        # context_id.
        if self._is_open_edx_studio(lti_request):
            context_id = self._get_mandatory_param(lti_request, "context_id")
            # generate an identifier in a normalized format, based on the context id
            # pylint: disable=too-many-function-args
            generated_id = hashlib.shake_128(context_id.encode("utf-8")).hexdigest(20)
            remote_user_id = f"openedxstudio-{generated_id}"

        logger.debug(
            "User %s (consumer = %s) authenticated from LTI request",
            remote_user_id,
            lti_consumer.slug,
        )
        user_model = get_user_model()
        try:
            user = user_model.objects.get(
                lti_consumer=lti_consumer, lti_remote_user_id=remote_user_id
            )
        except user_model.DoesNotExist:
            username = f"{remote_user_id}@{lti_consumer.slug:s}"

            user = user_model.objects.create_user(
                username,
                email=email,
                lti_consumer=lti_consumer,
                lti_remote_user_id=remote_user_id,
                public_username=public_username,
            )
            logger.debug("User %s created in database", username)

        if not user.is_active:
            logger.debug("User %s is not active", user.username)
            raise PermissionDenied()
        return user

    def _get_remote_user_id(self, lti_request: LTI):
        """
        Get the remote unique user identifier.
        It can be in different LTI parameters, depends on the LTI consumer.
        """
        user_id = self._get_lti_param_with_fallback(
            lti_request,
            [
                # OpenEdx
                "lis_person_sourcedid",
                # Moodle
                "ext_user_username",
                # Fallback to anonymous user_id
                "user_id",
            ],
        )
        if user_id:
            return user_id
        logger.debug("Unable to find remote user id in LTI request")
        raise PermissionDenied()

    def _get_public_username(self, lti_request: LTI):
        """
        Try to get a human-friendly remote user id in the LTI request.
        It tries different LTI parameters, because it depends on the LTI consumer.
        If not found, it returns ""
        """
        return self._get_lti_param_with_fallback(
            lti_request,
            [
                # OpenEdx
                "lis_person_sourcedid",
                # Moodle
                "ext_user_username",
            ],
            default_value="",
        )

    @staticmethod
    def _get_lti_param_with_fallback(
        lti_request: LTI, parameter_names: List[str], default_value=None
    ):
        """
        Depending on the LTI consumer, LTI parameters name can differ for the same value.
        This utility method allows to search for a value in the LTI request, by giving multiple
        parameter names, by order of preference.

        Args:
            lti_request: the LTI request
            parameter_names: The name of the LTI parameters to get, by order of preference
            default_value: value returned if none of these LTI parameters is found

        Returns:
            The value found in the LTI request, or default_value otherwise

        """
        for param in parameter_names:
            value = lti_request.get_param(param)
            if value:
                return value
        return default_value

    @staticmethod
    def _is_open_edx_studio(lti_request):
        """
        Detects if the LTI launch request is coming from OpenedX studio.
        """
        return (
            lti_request.get_param("user_id") == "student"
            and not lti_request.get_param("lis_person_contact_email_primary")
            and not lti_request.get_param("lis_person_sourcedid")
            and not lti_request.get_param("ext_user_username")
            and lti_request.roles == ["instructor"]
        )



"""
This module defines a ``PermissionHandler`` abstraction that allows to
implement filter or access logic related to forums.
"""
from typing import Optional

from machina.apps.forum_permission.handler import (
    PermissionHandler as BasePermissionHandler,
)


class PermissionHandler(BasePermissionHandler):
    """Add ashley specific permission checks to the machina permission handler."""

    def __init__(self):
        """Constructor"""
        super().__init__()
        # This attribute is used to store the LTIContext of the current user.
        # (see middleware.ForumPermission)
        # It allows the permission checking functions to be aware of the current LTIContext
        # we are scoped into, if any.
        self.current_lti_context_id: Optional[int] = None

    def can_rename_forum(self, forum, user):
        """ Given a forum, checks whether the user can rename it. """
        return self._perform_basic_permission_check(forum, user, "can_rename_forum")

    def can_manage_moderator(self, forum, user):
        """
        Given a forum, checks whether the user can promote or revoke a user to moderator.
        """
        return self._perform_basic_permission_check(forum, user, "can_manage_moderator")

    def forum_list_filter(self, qs, user):
        """
        Filters the given queryset in order to return a list of forums that
        can be seen and read by the specified user (at least).

        We override django machina's method to filter forums based on the
        current LTI context of the User, if any.
        """
        forums_to_show = super().forum_list_filter(qs, user)
        if self.current_lti_context_id:
            return forums_to_show.filter(lti_contexts__id=self.current_lti_context_id)
        return forums_to_show



"""
Template context processors
"""
import json

from django.conf import settings
from django.middleware.csrf import get_token


def site_metas(request):
    """Context processor to add all information required by Ashley templates and frontend."""
    if request.user.is_authenticated:
        return {
            "FRONTEND_CONTEXT": json.dumps(
                {
                    "context": {
                        "csrftoken": get_token(request),
                        "max_upload": settings.MAX_UPLOAD_FILE_MB,
                        "image_type": settings.IMAGE_TYPE_ALLOWED,
                    }
                }
            ),
        }
    return {}



"""Middleware declaration for the forum_permission machina app"""

from machina.apps.forum_permission.middleware import (
    ForumPermissionMiddleware as BaseForumPermissionMiddleware,
)

from ashley import SESSION_LTI_CONTEXT_ID


class ForumPermissionMiddleware(BaseForumPermissionMiddleware):
    """
    Django Machina's ForumPermissionMiddleware attaches an instance of the
    PermissionHandler to each request.

    We override it to store the current LTI context ID (that is stored in
    session), when it is available, in the permission handler.

    Since machina permission checking functions are not aware of the request
    object, this allows to use this information for permission checking.
    """

    def process_request(self, request):
        super().process_request(request)

        if request.user.is_authenticated:
            current_lti_context_id = request.session.get(SESSION_LTI_CONTEXT_ID)
            if current_lti_context_id:
                request.forum_permission_handler.current_lti_context_id = (
                    current_lti_context_id
                )



# -*- coding: utf-8 -*-
# Generated by Django 1.9.3 on 2016-03-28 02:35
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("forum", "0002_auto_20150725_0512"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="forum",
            name="is_active",
        ),
    ]



# Generated by Django 3.0.4 on 2020-03-25 19:12
import uuid

from django.db import migrations, models


def make_many_contexts(apps, schema_editor):
    """
    Create a LTIContext with Forum.consumer and Forum.lti_context fields
    stored in the Forum object, and add it to the many-to-many relationship
    in Forum.lti_contexts
    """
    Forum = apps.get_model("forum", "Forum")
    LTIContext = apps.get_model("ashley", "LTIContext")

    for forum in Forum.objects.all():
        lti_context, _created = LTIContext.objects.get_or_create(
            lti_id=forum.lti_context, lti_consumer=forum.lti_consumer
        )
        forum.lti_contexts.add(lti_context)


def gen_uuid(apps, schema_editor):
    Forum = apps.get_model("forum", "Forum")
    for forum in Forum.objects.all():
        forum.lti_id = uuid.uuid4()
        forum.save(update_fields=["lti_id"])


class Migration(migrations.Migration):

    dependencies = [
        ("ashley", "0002_auto_20200325_1651"),
        ("forum", "0012_auto_20200311_1354"),
    ]

    operations = [
        migrations.AddField(
            model_name="forum",
            name="lti_contexts",
            field=models.ManyToManyField(to="ashley.LTIContext"),
        ),
        migrations.RunPython(make_many_contexts),
        migrations.RemoveField(model_name="forum", name="lti_consumer"),
        migrations.RemoveField(model_name="forum", name="lti_context"),
        migrations.AddField(
            model_name="forum",
            name="lti_id",
            field=models.UUIDField(default=uuid.uuid4, null=True),
        ),
        migrations.RunPython(gen_uuid, reverse_code=migrations.RunPython.noop),
        migrations.AlterField(
            model_name="forum",
            name="lti_id",
            field=models.UUIDField(default=uuid.uuid4, editable=False, unique=True),
        ),
    ]



# Generated by Django 2.1.2 on 2018-10-24 02:44

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("forum_conversation", "0010_auto_20170120_0224"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="post",
            name="poster_ip",
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("forum", "0001_initial"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="forum",
            options={
                "ordering": ["tree_id", "lft"],
                "verbose_name": "Forum",
                "verbose_name_plural": "Forums",
            },
        ),
    ]



# Generated by Django 3.2.4 on 2021-06-23 12:56

import django.core.validators
import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models

import ashley.models
import ashley.validators


class Migration(migrations.Migration):

    dependencies = [
        ("forum", "0015_auto_20210217_1351"),
        ("ashley", "0004_auto_20201214_1750"),
    ]

    operations = [
        migrations.CreateModel(
            name="UploadImage",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "created",
                    models.DateTimeField(
                        auto_now_add=True, verbose_name="Creation date"
                    ),
                ),
                (
                    "updated",
                    models.DateTimeField(auto_now=True, verbose_name="Update date"),
                ),
                (
                    "file",
                    models.ImageField(
                        blank=True,
                        upload_to=ashley.models.image_directory_path,
                        validators=[
                            django.core.validators.FileExtensionValidator(
                                ["gif", "jpeg", "jpg", "png", "svg"]
                            ),
                            ashley.validators.validate_upload_image_file_size,
                        ],
                    ),
                ),
                (
                    "forum",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="forum.forum",
                        verbose_name="Forum",
                    ),
                ),
                (
                    "poster",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                        verbose_name="Poster",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
    ]



"""Ashley application"""
from machina.apps.forum_search.apps import (
    ForumSearchAppConfig as BaseForumSearchAppConfig,
)


class ForumSearchAppConfig(BaseForumSearchAppConfig):
    """Configuration class for the forum_search app."""

    name = "ashley.machina_extensions.forum_search"



"""Ashley application"""
from machina.apps.forum.apps import ForumAppConfig as BaseForumAppConfig


class ForumAppConfig(BaseForumAppConfig):
    """Configuration class for the forum app."""

    name = "ashley.machina_extensions.forum"



"""
This module defines the default permissions that can be configured for a forum
application.
"""

from django.utils.translation import gettext_lazy as _
from machina.apps.forum_permission.defaults import (
    PermissionConfig as BasePermissionConfig,
)


class PermissionConfig(BasePermissionConfig):
    """
    Overrides the PermissionConfig class from machina forum_permission
    application to add custom permissions.
    """

    permissions = BasePermissionConfig.permissions + [
        # Forums
        {
            "codename": "can_rename_forum",
            "label": _("Can rename a forum"),
            "scope": "forum",
        },
        {
            "codename": "can_manage_moderator",
            "label": _("Can manage a user's moderator status"),
            "scope": "forum",
        },
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.1 on 2017-05-24 02:40
from __future__ import unicode_literals

from django.db import migrations


def update_forum_last_post_on(apps, schema_editor):
    Forum = apps.get_model("forum", "Forum")
    Topic = apps.get_model("forum_conversation", "Topic")
    for forum in Forum.objects.all():
        topics = Topic.objects.filter(forum=forum, approved=True)
        forum.last_post_on = topics[0].last_post_on if topics.exists() else None
        forum.save()


class Migration(migrations.Migration):

    dependencies = [
        ("forum", "0006_auto_20170523_2036"),
    ]

    operations = [
        migrations.RunPython(
            update_forum_last_post_on, reverse_code=migrations.RunPython.noop
        ),
    ]



# Generated by Django 3.1.6 on 2021-02-17 13:51

import uuid

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("forum", "0014_auto_20210215_1542"),
    ]

    operations = [
        migrations.AlterField(
            model_name="forum",
            name="lti_id",
            field=models.UUIDField(db_index=True, default=uuid.uuid4, editable=False),
        ),
    ]



"""
    Forum views
    ===========

    This module defines views provided by the ``forum`` application.

"""
from django.urls import reverse
from django.utils.http import urlencode
from django.views.generic import UpdateView
from machina.apps.forum.views import ForumView as BaseForumView
from machina.apps.forum_permission.viewmixins import (
    PermissionRequiredMixin as BasePermissionRequiredMixin,
)
from machina.core.db.models import get_model
from machina.core.loading import get_class

Forum = get_model("forum", "Forum")

ORDER_VAR = "o"
PermissionRequiredMixin: BasePermissionRequiredMixin = get_class(
    "forum_permission.viewmixins", "PermissionRequiredMixin"
)


class ForumView(BaseForumView):  # pylint: disable=too-many-ancestors
    """Displays a forum and its topics."""

    # header columns
    list_display = ["subject", "posts_count", "views_count", "last_post_on"]
    # default ordering
    default_ordering = "-last_post_on"

    def get_default_index_order(self):
        """Utility to access index query column and column of default order with sign"""
        field = self.default_ordering.replace("-", "")
        sign = "-" if "-" in self.default_ordering else ""
        return {
            "index": f"{sign}{self.list_display.index(field)}",
            "col": self.default_ordering,
        }

    def __init__(self, **kwargs):
        super().__init__()
        self.params = {}

    def get(self, request, **kwargs):
        """Handles GET requests."""
        # collect the current params for the order by in request
        self.params = dict(request.GET.items())
        response = super().get(request, **kwargs)
        return response

    def get_ordering(self):
        """Gets order requested in url"""
        ordering = self.request.GET.get(
            ORDER_VAR, self.get_default_index_order().get("index")
        )
        try:
            _none, pfx, idx = ordering.rpartition("-")
            # returns default column if out of index order
            if (int)(idx) >= len(self.list_display):
                idx = self.get_default_index_order().get("col")
            return pfx, (int)(idx)
        except ValueError:
            # param exists but value is unexpected, use default order
            _none, pfx, idx = (
                self.get_default_index_order().get("index").rpartition("-")
            )
            return pfx, (int)(idx)

    def get_queryset(self):
        """Returns the list of items for this view ordered by asked param."""
        query = super().get_queryset()
        # Type of topic is kept as first order argument to keep sticky option
        return query.order_by("-type", self.get_ordering_column())

    def get_context_data(self, **kwargs):
        """Returns the context data to provide to the template."""
        context = super().get_context_data(**kwargs)
        # add sort header links containing asc, desc orders
        context["header"] = list(self.result_headers())

        # used for pagination to keep the order, make sure it has expected value
        if self.request.GET.get(ORDER_VAR):
            pfx, idx = self.get_ordering()
            # if value is unexpected, order is ommited
            if f"{pfx}{idx}" == self.request.GET.get(ORDER_VAR):
                context["order"] = f"{ORDER_VAR}={pfx}{idx}"

        return context

    def result_headers(self):
        """Builds each link and css class attribute for the header depending on current order"""
        pfx, idx = self.get_ordering()

        for index, _field_name in enumerate(self.list_display):
            new_order_type = "asc"
            div_classes = ""
            query_string = self.get_query_string({ORDER_VAR: "-" + str(index)})
            # Is it currently being sorted on?
            is_sorted = index == idx
            if is_sorted:
                order_type = "desc" if pfx == "-" else "asc"
                div_classes = f"sorted {order_type}ending"
                new_order_type = {"asc": "desc", "desc": "asc"}[order_type]
                query_string = self.get_query_string(
                    {ORDER_VAR: ("-" if pfx == "" else "") + str(index)}
                )

            yield {
                "sorted": is_sorted,
                "ascending": new_order_type == "desc",
                "url_order": query_string,
                "class_attrib": div_classes,
            }

    def get_query_string(self, new_params):
        """
        Build the query_string for this link keeping all the query params and replacing
        or adding new_params
        """
        param = {**self.params, **new_params}
        return "?%s" % urlencode(sorted(param.items()))

    def get_ordering_column(self):
        """From query url, retrieve the concerned field ordered on if any."""
        pfx, idx = self.get_ordering()
        if idx < len(self.list_display) and self.list_display[idx]:
            return pfx + self.list_display[idx]

        return self.get_default_index_order().get("col")


class ForumRenameView(PermissionRequiredMixin, UpdateView):
    """Allows the current user to update its forum profile."""

    model = Forum
    fields = ["name"]
    template_name = "forum/forum_rename.html"

    def get_success_url(self):
        """Returns the success URL to redirect the user to."""
        return reverse(
            "forum:forum",
            kwargs={
                "pk": self.object.pk,
                "slug": self.object.slug,
            },
        )

    # pylint: disable=unused-argument
    def perform_permissions_check(self, user, obj, perms):
        """Performs the permissions check."""
        return self.request.forum_permission_handler.can_rename_forum(obj, user)



"""Declare the models related to Ashley ."""
import uuid

from django.db import models
from machina.apps.forum.abstract_models import AbstractForum as MachinaAbstractForum
from machina.core.db.models import get_model, model_factory

LTIContext = get_model("ashley", "LTIContext")  # pylint: disable=C0103


class AbstractForum(MachinaAbstractForum):
    """
    Forum model for Ashley.

    It is based on django-machina's Forum model and adds fields to be able to
    map a LTI request to a forum.
    """

    # pylint: disable=all
    # Pylint is disabled because it crashes while testing Foreign keys declared in the
    # django-machina's parent abstract model. This is a known issue in pylint-django with
    # foreign keys models referenced by their name.
    # (See https://github.com/PyCQA/pylint-django#known-issues )
    lti_id = models.UUIDField(
        null=False, default=uuid.uuid4, editable=False, unique=False, db_index=True
    )

    lti_contexts = models.ManyToManyField(LTIContext)

    class Meta(MachinaAbstractForum.Meta):
        abstract = True


Forum = model_factory(AbstractForum)

from machina.apps.forum.abstract_models import *  # noqa isort:skip



# Generated by Django 2.1.2 on 2018-11-03 19:01

import machina.apps.forum.abstract_models
import machina.models.fields
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("forum", "0009_auto_20170928_2327"),
    ]

    operations = [
        migrations.AlterField(
            model_name="forum",
            name="image",
            field=machina.models.fields.ExtendedImageField(
                blank=True,
                null=True,
                upload_to=machina.apps.forum.abstract_models.get_forum_image_upload_to,
                verbose_name="Forum image",
            ),
        ),
    ]



# Generated by Django 3.0.4 on 2020-03-25 16:51

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("lti_toolbox", "0001_initial"),
        ("ashley", "0001_initial"),
    ]

    operations = [
        migrations.CreateModel(
            name="LTIContext",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("lti_id", models.CharField(db_index=True, max_length=150)),
                (
                    "lti_consumer",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="lti_toolbox.LTIConsumer",
                    ),
                ),
            ],
            options={"abstract": False},
        ),
        migrations.AddConstraint(
            model_name="lticontext",
            constraint=models.UniqueConstraint(
                fields=("lti_consumer", "id"),
                name="lticontext_unique_consumer_context_id",
            ),
        ),
    ]



"""
Import the admin classes from the django-machina's forum_permission application
"""

# pylint: disable=unused-wildcard-import,wildcard-import
from machina.apps.forum_permission.admin import *  # noqa isort:skip



"""
Ashley URLs (that includes django machina urls)
"""

from django.urls import include, path, re_path
from machina import urls as machina_urls

from ashley.api import urls as api_urls
from ashley.views import ChangeUsernameView, ForumLTIView, ManageModeratorsView

API_PREFIX = "v1.0"

urlpatterns = [
    path("lti/forum/<uuid:uuid>", ForumLTIView.as_view(), name="forum.lti.view"),
    path(
        "profile/username",
        ChangeUsernameView.as_view(),
        name="forum.username.change",
    ),
    path("moderators/", ManageModeratorsView.as_view(), name="moderators"),
    re_path(r"api/{}/".format(API_PREFIX), include(api_urls)),
    path("forum/", include(machina_urls)),
]



"""
    Forum search forms
    ==================
    This module defines forms provided by the ``forum_search`` application.
"""
from machina.apps.forum_search.forms import SearchForm as MachinaSearchForm
from machina.core.db.models import get_model
from machina.core.loading import get_class

Forum = get_model("forum", "Forum")
PermissionHandler = get_class("forum_permission.handler", "PermissionHandler")


class SearchForm(MachinaSearchForm):
    """
    Override Django Machina's SearchForm to allow searching poster user names
    even with an empty search in the main field.
    """

    def __init__(self, *args, **kwargs):
        """
        Init form, code based on base class MachinaSearchForm. A filter on lti_context
        has been added.
        """
        # Loads current lti_context to filter forum search
        lti_contexts = kwargs.pop("lti_context", None)
        super().__init__(*args, **kwargs)
        user = kwargs.pop("user", None)
        self.allowed_forums = PermissionHandler().get_readable_forums(
            Forum.objects.filter(lti_contexts=lti_contexts), user
        )
        # self.allowed_forums is used in search method of MachinaSearchForm
        if self.allowed_forums:
            self.fields["search_forums"].choices = [
                (f.id, "{} {}".format("-" * f.margin_level, f.name))
                for f in self.allowed_forums
            ]

    def clean(self):
        """
        Set main query to catch all "*" if it is empty and there is a search term for
        poster username.
        """
        cleaned_data = super().clean()

        if cleaned_data.get("search_poster_name") and not cleaned_data.get("q"):
            cleaned_data["q"] = "*"

        return cleaned_data



"""Factories for the ``ashley`` application."""
import json
import uuid

import factory
from django.contrib.auth import get_user_model
from django.utils.text import slugify
from factory.django import DjangoModelFactory
from faker import Faker
from lti_toolbox.factories import LTIConsumerFactory
from machina.core.db.models import get_model

faker = Faker()

Forum = get_model("forum", "Forum")  # pylint: disable=C0103
LTIContext = get_model("ashley", "LTIContext")  # pylint: disable=C0103
Post = get_model("forum_conversation", "Post")  # pylint: disable=C0103
Topic = get_model("forum_conversation", "Topic")  # pylint: disable=C0103
User = get_user_model()
UploadImage = get_model("ashley", "UploadImage")  # pylint: disable=C0103


class UserFactory(DjangoModelFactory):
    """Factory to create a User."""

    class Meta:
        model = User

    email = factory.LazyAttribute(lambda u: f"{u.public_username}@example.com")
    lti_consumer = factory.SubFactory(LTIConsumerFactory)
    lti_remote_user_id = factory.SelfAttribute("public_username")
    public_username = factory.Sequence(lambda n: f"user{n}")
    username = factory.LazyAttribute(
        lambda u: f"{u.public_username}@{u.lti_consumer.slug}"
    )


class LTIContextFactory(DjangoModelFactory):
    """Factory to create a LTIContext."""

    class Meta:
        model = LTIContext

    lti_consumer = factory.SubFactory(LTIConsumerFactory)
    lti_id = factory.Sequence(lambda n: f"context{n}")


class ForumFactory(DjangoModelFactory):
    """Factory to create a Forum"""

    class Meta:
        model = Forum

    name = factory.Sequence(lambda n: f"forum{n}")
    type = Forum.FORUM_POST


class TopicFactory(DjangoModelFactory):
    """Factory to create a Topic"""

    class Meta:
        model = Topic

    forum = factory.SubFactory(ForumFactory)
    poster = factory.SubFactory(UserFactory)
    slug = factory.LazyAttribute(lambda t: slugify(t.subject))
    status = Topic.TOPIC_UNLOCKED
    # Watchout: Django Machina is overriding the topic subject when saving the first post!
    # https://github.com/ellmetha/django-machina/blob/master/machina/apps/forum_conversation/abstract_models.py#L314
    subject = factory.Faker("text", max_nb_chars=100)
    type = Topic.TOPIC_POST


class PostFactory(DjangoModelFactory):
    """Factory to create a Post"""

    class Meta:
        model = Post
        exclude = ("text",)

    poster = factory.SubFactory(UserFactory)
    subject = factory.Faker("text", max_nb_chars=100)
    text = factory.Faker("paragraph", nb_sentences=5)
    topic = factory.SubFactory(TopicFactory)

    @factory.lazy_attribute
    def content(self):
        """Building a fake content around the text as would be generated by draftjs."""
        data = {
            "blocks": [
                {
                    "key": faker.uuid4(),
                    "text": self.text,
                    "type": "unstyled",
                    "depth": 0,
                    "inlineStyleRanges": [],
                    "entityRanges": [],
                    "data": {},
                }
            ],
            "entityMap": {},
        }
        return json.dumps(data)


class UploadImageFactory(DjangoModelFactory):
    """Factory to create an UploadImage"""

    class Meta:
        model = UploadImage

    forum = factory.SubFactory(ForumFactory)
    poster = factory.SubFactory(UserFactory)
    file = f"{uuid.uuid4()}.png"



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("forum", "0001_initial"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("auth", "0001_initial"),
    ]

    operations = [
        migrations.CreateModel(
            name="ForumPermission",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                (
                    "codename",
                    models.CharField(
                        unique=True, max_length=150, verbose_name="Permission codename"
                    ),
                ),
                (
                    "name",
                    models.CharField(
                        max_length=255,
                        null=True,
                        verbose_name="Permission name",
                        blank=True,
                    ),
                ),
                (
                    "is_global",
                    models.BooleanField(
                        default=False,
                        help_text="This permission can be granted globally to all the forums",
                        verbose_name="Global permission",
                    ),
                ),
                (
                    "is_local",
                    models.BooleanField(
                        default=True,
                        help_text="This permission can be granted individually for each forum",
                        verbose_name="Local permission",
                    ),
                ),
            ],
            options={
                "abstract": False,
                "verbose_name": "Forum permission",
                "verbose_name_plural": "Forum permissions",
            },
        ),
        migrations.CreateModel(
            name="GroupForumPermission",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                (
                    "has_perm",
                    models.BooleanField(default=True, verbose_name="Has perm"),
                ),
                (
                    "forum",
                    models.ForeignKey(
                        verbose_name="Forum",
                        blank=True,
                        to="forum.Forum",
                        null=True,
                        on_delete=models.CASCADE,
                    ),
                ),
                (
                    "group",
                    models.ForeignKey(
                        verbose_name="Group", to="auth.Group", on_delete=models.CASCADE
                    ),
                ),
                (
                    "permission",
                    models.ForeignKey(
                        verbose_name="Forum permission",
                        to="forum_permission.ForumPermission",
                        on_delete=models.CASCADE,
                    ),
                ),
            ],
            options={
                "abstract": False,
                "verbose_name": "Group forum permission",
                "verbose_name_plural": "Group forum permissions",
            },
        ),
        migrations.CreateModel(
            name="UserForumPermission",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                (
                    "has_perm",
                    models.BooleanField(default=True, verbose_name="Has perm"),
                ),
                (
                    "anonymous_user",
                    models.BooleanField(
                        default=False, verbose_name="Target anonymous user"
                    ),
                ),
                (
                    "forum",
                    models.ForeignKey(
                        verbose_name="Forum",
                        blank=True,
                        to="forum.Forum",
                        null=True,
                        on_delete=models.CASCADE,
                    ),
                ),
                (
                    "permission",
                    models.ForeignKey(
                        verbose_name="Forum permission",
                        to="forum_permission.ForumPermission",
                        on_delete=models.CASCADE,
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        verbose_name="User",
                        blank=True,
                        to=settings.AUTH_USER_MODEL,
                        null=True,
                        on_delete=models.CASCADE,
                    ),
                ),
            ],
            options={
                "abstract": False,
                "verbose_name": "User forum permission",
                "verbose_name_plural": "User forum permissions",
            },
        ),
        migrations.AlterUniqueTogether(
            name="userforumpermission",
            unique_together=set([("permission", "forum", "user")]),
        ),
        migrations.AlterUniqueTogether(
            name="groupforumpermission",
            unique_together=set([("permission", "forum", "group")]),
        ),
    ]



"""Ashley template tags to display forum content"""
from django import template
from machina.core.db.models import get_model

from ashley.defaults import _FORUM_ROLE_INSTRUCTOR

LTIContext = get_model("ashley", "LTIContext")
Topic = get_model("forum_conversation", "Topic")

register = template.Library()


@register.filter()
def is_user_instructor(topic, user):
    """
    This will return a boolean indicating if the passed user is instructor
    for the given topic.
    Usage::
        {% if topic|is_user_instructor:user %}...{% endif %}
    """
    # get the list of lti concerns by this forum, one forum can have multiple LTI Context
    # if the user is instructor in one of them then he is considered
    # as instructor for this forum
    return any(
        _FORUM_ROLE_INSTRUCTOR in lti.get_user_roles(user)
        for lti in LTIContext.objects.filter(forum=topic.forum)
    )



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import machina.models.fields
import mptt.fields
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Forum",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                ("is_active", models.BooleanField(default=True, db_index=True)),
                (
                    "created",
                    models.DateTimeField(
                        auto_now_add=True, verbose_name="Creation date"
                    ),
                ),
                (
                    "updated",
                    models.DateTimeField(auto_now=True, verbose_name="Update date"),
                ),
                ("name", models.CharField(max_length=100, verbose_name="Name")),
                ("slug", models.SlugField(max_length=255, verbose_name="Slug")),
                (
                    "description",
                    machina.models.fields.MarkupTextField(
                        no_rendered_field=True,
                        null=True,
                        verbose_name="Description",
                        blank=True,
                    ),
                ),
                (
                    "image",
                    machina.models.fields.ExtendedImageField(
                        upload_to="machina/forum_images",
                        null=True,
                        verbose_name="Forum image",
                        blank=True,
                    ),
                ),
                (
                    "link",
                    models.URLField(null=True, verbose_name="Forum link", blank=True),
                ),
                (
                    "link_redirects",
                    models.BooleanField(
                        default=False,
                        help_text="Records the number of times a forum link was clicked",
                        verbose_name="Track link redirects count",
                    ),
                ),
                (
                    "type",
                    models.PositiveSmallIntegerField(
                        db_index=True,
                        verbose_name="Forum type",
                        choices=[
                            (0, "Default forum"),
                            (1, "Category forum"),
                            (2, "Link forum"),
                        ],
                    ),
                ),
                (
                    "posts_count",
                    models.PositiveIntegerField(
                        default=0,
                        verbose_name="Number of posts",
                        editable=False,
                        blank=True,
                    ),
                ),
                (
                    "topics_count",
                    models.PositiveIntegerField(
                        default=0,
                        verbose_name="Number of topics",
                        editable=False,
                        blank=True,
                    ),
                ),
                (
                    "link_redirects_count",
                    models.PositiveIntegerField(
                        default=0,
                        verbose_name="Track link redirects count",
                        editable=False,
                        blank=True,
                    ),
                ),
                (
                    "last_post_on",
                    models.DateTimeField(
                        null=True, verbose_name="Last post added on", blank=True
                    ),
                ),
                (
                    "display_sub_forum_list",
                    models.BooleanField(
                        default=True,
                        help_text="Displays this forum on the legend of its parent-forum (sub forums list)",
                        verbose_name="Display in parent-forums legend",
                    ),
                ),
                (
                    "_description_rendered",
                    models.TextField(null=True, editable=False, blank=True),
                ),
                ("lft", models.PositiveIntegerField(editable=False, db_index=True)),
                ("rght", models.PositiveIntegerField(editable=False, db_index=True)),
                ("tree_id", models.PositiveIntegerField(editable=False, db_index=True)),
                ("level", models.PositiveIntegerField(editable=False, db_index=True)),
                (
                    "parent",
                    mptt.fields.TreeForeignKey(
                        related_name="children",
                        verbose_name="Parent",
                        blank=True,
                        to="forum.Forum",
                        null=True,
                        on_delete=models.CASCADE,
                    ),
                ),
            ],
            options={
                "ordering": ["tree_id", "lft"],
                "abstract": False,
                "verbose_name": "Forum",
                "verbose_name_plural": "Forums",
                "permissions": [],
            },
        ),
    ]



"""This application overrides to django-machina's forum_conversation app."""

# pylint: disable=invalid-name
default_app_config = (
    "ashley.machina_extensions.forum_conversation.apps.ForumConversationAppConfig"
)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.5 on 2017-09-29 04:27
from __future__ import unicode_literals

from django.db import migrations


def update_forum_last_topic_and_last_post(apps, schema_editor):
    Forum = apps.get_model("forum", "Forum")
    Topic = apps.get_model("forum_conversation", "Topic")
    for forum in Forum.objects.all():
        topics = Topic.objects.filter(forum=forum, approved=True)
        if topics.exists():
            forum.last_post_id = topics[0].last_post_id
            forum.save()


class Migration(migrations.Migration):

    dependencies = [
        ("forum", "0008_forum_last_post"),
    ]

    operations = [
        migrations.RunPython(
            update_forum_last_topic_and_last_post,
            reverse_code=migrations.RunPython.noop,
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11 on 2017-05-05 02:13
from __future__ import unicode_literals

from django.db import migrations


def update_forum_direct_counts(apps, schema_editor):
    Forum = apps.get_model("forum", "Forum")
    Topic = apps.get_model("forum_conversation", "Topic")
    for forum in Forum.objects.all():
        topics = Topic.objects.filter(forum=forum, approved=True)
        forum.direct_topics_count = topics.count()
        forum.direct_posts_count = sum(t.posts_count for t in topics)
        forum.save()


class Migration(migrations.Migration):

    dependencies = [
        ("forum", "0004_auto_20170504_2108"),
        ("forum_conversation", "0010_auto_20170120_0224"),
    ]

    operations = [
        migrations.RunPython(
            update_forum_direct_counts, reverse_code=migrations.RunPython.noop
        ),
    ]



"""Validators for the ashley application."""
from html.parser import HTMLParser

from django.conf import settings
from django.core import validators
from django.core.exceptions import ValidationError
from django.utils.translation import gettext_lazy as _
from machina.models.fields import render_func


class MarkupNullableMaxLengthValidator(validators.MaxLengthValidator):
    """
    Ensure that the length of the plain text contained in a markup is lower than
    or equal the specified max length. Also, it provides a way to not validate
    an input if the max length is None.
    """

    def __call__(self, value):
        if self.limit_value is None:
            # If the limit value is None, this means that there is no
            # limit value at all. The default validation process is not
            # performed.
            return
        # Render the value to HTML using MACHINA_MARKUP_LANGUAGE
        html_value = render_func(value)
        # Extract plain text from html
        html_parser = HTMLFilter()
        html_parser.feed(html_value)
        # Validate the extracted text length
        super().__call__(html_parser.text)


class HTMLFilter(HTMLParser):
    """HTML parser to extract plain text from a HTML fragment."""

    text = ""

    def error(self, message):
        """Ignore errors"""

    def handle_data(self, data):
        """Concatenate the plain text found in each element."""
        self.text += data


def validate_upload_image_file_size(file):
    """Controls the size of the uploaded file is not over the limit authorized"""
    if file.size > settings.MAX_UPLOAD_FILE_MB * 1024 * 1024:
        raise ValidationError(
            _(
                f"The maximum file size that can be uploaded is {settings.MAX_UPLOAD_FILE_MB} MB"
            )
        )
    return file



"""
    Forum conversation forms
    ========================

    This module defines forms provided by the ``forum_conversation`` application.

"""
from machina.apps.forum_conversation.forms import PostForm as MachinaPostForm
from machina.apps.forum_conversation.forms import TopicForm as MachinaTopicForm


class PostForm(MachinaPostForm):
    """ Overload Machina PostForm to send extra variables to the widget editor """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # collect extra informations used by the editor
        self.fields["content"].widget.attrs.update(
            {
                "mentions": self.topic.get_active_users(self.user),
                "forum": self.forum.id,
            }
        )

        # remove unused machina placeholder, placeholder is created in our component AshleyEditor
        self.fields["content"].widget.attrs.pop("placeholder")


class TopicForm(MachinaTopicForm):
    """Overload Machina TopicForm to send extra variables to the widget editor for the
    Topic creation. TopicForm is loaded only for the initial post of the topic."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.topic is not None and self.topic.posts_count > 1:
            # will be useful only if this first post is edited after answers
            self.fields["content"].widget.attrs[
                "mentions"
            ] = self.topic.get_active_users(self.user)

        # send extra informations used by the editor
        self.fields["content"].widget.attrs.update({"forum": self.forum.id})

        # remove unused machina placeholder, placeholder is created in our component AshleyEditor
        self.fields["content"].widget.attrs.pop("placeholder")



# -*- coding: utf-8 -*-
# Generated by Django 1.11 on 2017-05-05 02:08
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("forum", "0003_remove_forum_is_active"),
    ]

    operations = [
        migrations.AddField(
            model_name="forum",
            name="direct_posts_count",
            field=models.PositiveIntegerField(
                blank=True,
                default=0,
                editable=False,
                verbose_name="Direct number of posts",
            ),
        ),
        migrations.AddField(
            model_name="forum",
            name="direct_topics_count",
            field=models.PositiveIntegerField(
                blank=True,
                default=0,
                editable=False,
                verbose_name="Direct number of topics",
            ),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.2 on 2016-02-28 19:51
from __future__ import unicode_literals

import machina.core.validators
import machina.models.fields
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("forum_conversation", "0002_post_anonymous_key"),
    ]

    operations = [
        migrations.AlterField(
            model_name="post",
            name="content",
            field=machina.models.fields.MarkupTextField(
                no_rendered_field=True,
                validators=[machina.core.validators.NullableMaxLengthValidator(None)],
                verbose_name="Content",
            ),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.10.4 on 2017-01-20 01:24
from __future__ import unicode_literals

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("forum_conversation", "0009_auto_20160925_2126"),
    ]

    operations = [
        migrations.AlterField(
            model_name="topic",
            name="first_post",
            field=models.ForeignKey(
                blank=True,
                editable=False,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="+",
                to="forum_conversation.Post",
                verbose_name="First post",
            ),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.3 on 2016-04-27 03:02
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("forum_conversation", "0003_auto_20160228_2051"),
    ]

    operations = [
        migrations.AlterField(
            model_name="topic",
            name="subscribers",
            field=models.ManyToManyField(
                blank=True,
                related_name="topic_subscriptions",
                to=settings.AUTH_USER_MODEL,
                verbose_name="Subscribers",
            ),
        ),
    ]



"""Declare the models related to the forum_permission app ."""

# Import models from the django-machina app
from machina.apps.forum_permission.models import *  # noqa, pylint: disable=unused-wildcard-import,wildcard-import



"""Ashley application"""
from machina.apps.forum_conversation.apps import (
    ForumConversationAppConfig as BaseForumConversationAppConfig,
)


class ForumConversationAppConfig(BaseForumConversationAppConfig):
    """Configuration class for the forum_conversation app."""

    name = "ashley.machina_extensions.forum_conversation"



# -*- coding: utf-8 -*-
# Generated by Django 1.9.6 on 2016-06-07 02:55
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("forum_conversation", "0004_auto_20160427_0502"),
    ]

    operations = [
        migrations.AlterField(
            model_name="post",
            name="approved",
            field=models.BooleanField(
                db_index=True, default=True, verbose_name="Approved"
            ),
        ),
        migrations.AlterField(
            model_name="topic",
            name="approved",
            field=models.BooleanField(
                db_index=True, default=True, verbose_name="Approved"
            ),
        ),
    ]



"""
This module provides a management command `sync_group_permissions`
to help ashley operators maintaining a consistent set of forum permission
across all LTI related groups.
"""

from django.core.management.base import BaseCommand
from machina.apps.forum_permission.shortcuts import assign_perm, remove_perm
from machina.core.db.models import get_model

from ashley.defaults import DEFAULT_FORUM_ROLES_PERMISSIONS

Forum = get_model("forum", "Forum")  # pylint: disable=C0103
GroupForumPermission = get_model(
    "forum_permission", "GroupForumPermission"
)  # pylint: disable=C0103
LTIContext = get_model("ashley", "LTIContext")  # pylint: disable=C0103


class Command(BaseCommand):
    """
    Implementation of the sync_group_permission Command.
    """

    help = (
        "Update group permissions in database with permissions defined "
        "in ASHLEY_DEFAULT_FORUM_ROLES_PERMISSIONS. By default, only missing "
        "permissions are added and no permission is removed."
    )
    apply_updates = False
    remove_extra_permissions = False

    def add_arguments(self, parser):
        """Set custom arguments for this command."""

        parser.add_argument(
            "--apply",
            action="store_true",
            help="Apply the update in database (by default, the command runs in dry mode)",
        )
        parser.add_argument(
            "--remove-extra-permissions",
            action="store_true",
            help="Revoke group permissions that are not defined in the settings",
        )

    def handle(self, *args, **options):
        """Command handler, that execute the actual logic of the command."""

        self.apply_updates = bool(options["apply"])
        self.remove_extra_permissions = bool(options["remove_extra_permissions"])

        self.stdout.write("Synchronizing group permissions")
        if not self.apply_updates:
            self.stdout.write("(DRY RUN)")

        for forum in Forum.objects.iterator():
            self.sync_forum_permissions(forum)

        self.stdout.write("Done.")

    def sync_forum_permissions(self, forum):
        """Synchronize permissions for every LTI role group related to a forum."""
        for lti_context in forum.lti_contexts.iterator():
            # pylint: disable=no-member
            for role, permissions in DEFAULT_FORUM_ROLES_PERMISSIONS.items():
                role_group = lti_context.get_role_group(role)
                self.sync_forum_group_permissions(forum, role_group, permissions)

    def sync_forum_group_permissions(self, forum, group, expected_permissions):
        """Set permissions for a group in a specific forum"""
        current_permissions = GroupForumPermission.objects.filter(
            forum=forum, group=group, has_perm=True
        ).values_list("permission__codename", flat=True)

        # Add missing permissions
        for permission in expected_permissions:
            if permission not in current_permissions:
                self.grant_permission(permission, group, forum)
        # Remove extra permissions
        if self.remove_extra_permissions:
            for permission in current_permissions:
                if permission not in expected_permissions:
                    self.revoke_permission(permission, group, forum)

    def grant_permission(self, permission, group, forum):
        """Grant a permission to a group in the specified forum."""
        if self.apply_updates:
            assign_perm(permission, group, forum, True)
        self.stdout.write(
            f"ADDED {permission} for group {group.name} in forum {forum.pk} ({forum.slug})"
        )

    def revoke_permission(self, permission, group, forum):
        """Revoke a permission from a group in the specified forum."""
        if self.apply_updates:
            remove_perm(permission, group, forum)
        self.stdout.write(
            f"REMOVED {permission} for group {group.name} in forum {forum.pk} ({forum.slug})"
        )



# Generated by Django 2.1.7 on 2019-03-20 03:40

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("forum_permission", "0003_remove_forumpermission_name"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="forumpermission",
            name="is_global",
        ),
        migrations.RemoveField(
            model_name="forumpermission",
            name="is_local",
        ),
    ]



"""Ashley API URLs"""
from django.urls import include, path
from rest_framework.routers import DefaultRouter

from .views import ImageUploadApiView, UserApiView

router = DefaultRouter()
router.register(
    "users",
    UserApiView,
    basename="users",
)
router.register(
    "images",
    ImageUploadApiView,
    basename="images",
)

urlpatterns = [
    path("", include(router.urls)),
]



# Generated by Django 2.2.10 on 2020-04-23 15:49

import machina.core.validators
import machina.models.fields
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("forum_conversation", "0011_remove_post_poster_ip"),
    ]

    operations = [
        migrations.AlterField(
            model_name="post",
            name="content",
            field=machina.models.fields.MarkupTextField(
                no_rendered_field=True,
                validators=[machina.core.validators.MarkupMaxLengthValidator(None)],
                verbose_name="Content",
            ),
        ),
    ]



"""
This package contains utilities for Ashley's WYSIWYG editor.
"""

import json
import logging

from draftjs_exporter.html import HTML

from ..defaults import DRAFTJS_EXPORTER_CONFIG

logger = logging.getLogger(__name__)

DRAFTJS_EXPORTER = HTML(DRAFTJS_EXPORTER_CONFIG)


def draftjs_renderer(content: str) -> str:
    """Convert a Draft.js JSON state into HTML"""
    try:
        return str(DRAFTJS_EXPORTER.render(json.loads(content)))
    except json.JSONDecodeError as err:
        logger.warning("Unable to render content : %s (content : [%s])", err, content)
    return ""



# Generated by Django 2.2.11 on 2020-03-10 09:34

import django.contrib.auth.models
import django.contrib.auth.validators
import django.db.models.deletion
import django.utils.timezone
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ("auth", "0011_update_proxy_permissions"),
        ("lti_toolbox", "0001_initial"),
    ]

    operations = [
        migrations.CreateModel(
            name="User",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("password", models.CharField(max_length=128, verbose_name="password")),
                (
                    "last_login",
                    models.DateTimeField(
                        blank=True, null=True, verbose_name="last login"
                    ),
                ),
                (
                    "is_superuser",
                    models.BooleanField(
                        default=False,
                        help_text="Designates that this user has all permissions without explicitly assigning them.",
                        verbose_name="superuser status",
                    ),
                ),
                (
                    "username",
                    models.CharField(
                        error_messages={
                            "unique": "A user with that username already exists."
                        },
                        help_text="Required. 150 characters or fewer. Letters, digits and @/./+/-/_ only.",
                        max_length=150,
                        unique=True,
                        validators=[
                            django.contrib.auth.validators.UnicodeUsernameValidator()
                        ],
                        verbose_name="username",
                    ),
                ),
                (
                    "first_name",
                    models.CharField(
                        blank=True, max_length=30, verbose_name="first name"
                    ),
                ),
                (
                    "last_name",
                    models.CharField(
                        blank=True, max_length=150, verbose_name="last name"
                    ),
                ),
                (
                    "email",
                    models.EmailField(
                        blank=True, max_length=254, verbose_name="email address"
                    ),
                ),
                (
                    "is_staff",
                    models.BooleanField(
                        default=False,
                        help_text="Designates whether the user can log into this admin site.",
                        verbose_name="staff status",
                    ),
                ),
                (
                    "is_active",
                    models.BooleanField(
                        default=True,
                        help_text="Designates whether this user should be treated as active. Unselect this instead of deleting accounts.",
                        verbose_name="active",
                    ),
                ),
                (
                    "date_joined",
                    models.DateTimeField(
                        default=django.utils.timezone.now, verbose_name="date joined"
                    ),
                ),
                (
                    "lti_remote_user_id",
                    models.CharField(
                        help_text="Unique identifier for the user on the tool consumer",
                        max_length=150,
                        verbose_name="LTI remote user identifier",
                    ),
                ),
                (
                    "public_username",
                    models.CharField(
                        help_text="This username will be displayed with the user's posts",
                        max_length=150,
                        verbose_name="Public username",
                    ),
                ),
                (
                    "groups",
                    models.ManyToManyField(
                        blank=True,
                        help_text="The groups this user belongs to. A user will get all permissions granted to each of their groups.",
                        related_name="user_set",
                        related_query_name="user",
                        to="auth.Group",
                        verbose_name="groups",
                    ),
                ),
                (
                    "lti_consumer",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="lti_toolbox.LTIConsumer",
                    ),
                ),
                (
                    "user_permissions",
                    models.ManyToManyField(
                        blank=True,
                        help_text="Specific permissions for this user.",
                        related_name="user_set",
                        related_query_name="user",
                        to="auth.Permission",
                        verbose_name="user permissions",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
            managers=[
                ("objects", django.contrib.auth.models.UserManager()),
            ],
        ),
        migrations.AddConstraint(
            model_name="user",
            constraint=models.UniqueConstraint(
                fields=("lti_consumer", "lti_remote_user_id"),
                name="ashleyuser_unique_consumer_remote_user_id",
            ),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11 on 2017-04-21 02:31
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("forum_permission", "0002_auto_20160607_0500"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="forumpermission",
            name="name",
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.7 on 2016-07-09 04:16
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("forum_conversation", "0005_auto_20160607_0455"),
    ]

    operations = [
        migrations.AddField(
            model_name="post",
            name="enable_signature",
            field=models.BooleanField(
                db_index=True, default=True, verbose_name="Attach a signature"
            ),
        ),
    ]



"""Factories for the ``forum`` application."""
import factory
from factory.django import DjangoModelFactory
from lti_toolbox.factories import LTIConsumerFactory
from machina.core.db.models import get_model

Forum = get_model("forum", "Forum")  # pylint: disable=C0103


class CourseForumFactory(DjangoModelFactory):
    """Factory to create a Forum related to a course."""

    class Meta:
        model = Forum

    type = Forum.FORUM_POST
    name = factory.Sequence(lambda n: f"Forum {n}")
    lti_context = factory.Sequence(lambda n: f"course{n}")
    lti_consumer = factory.SubFactory(LTIConsumerFactory)



# Generated by Django 3.1.1 on 2020-09-17 13:35

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("lti_toolbox", "0001_initial"),
        ("ashley", "0002_auto_20200325_1651"),
    ]

    operations = [
        migrations.AlterField(
            model_name="lticontext",
            name="lti_consumer",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                to="lti_toolbox.lticonsumer",
            ),
        ),
        migrations.AlterField(
            model_name="user",
            name="first_name",
            field=models.CharField(
                blank=True, max_length=150, verbose_name="first name"
            ),
        ),
        migrations.AlterField(
            model_name="user",
            name="lti_consumer",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                to="lti_toolbox.lticonsumer",
            ),
        ),
    ]



"""Mixins based on current LTIContext Session"""
from machina.core.db.models import get_model

from . import SESSION_LTI_CONTEXT_ID

LTIContext = get_model("ashley", "LTIContext")


def get_current_lti_session(request):
    """
    Gets from current session the corresponding LTIContext object.
    """
    if request.user.is_authenticated and request.session.get(SESSION_LTI_CONTEXT_ID):
        return LTIContext.objects.get(id=request.session[SESSION_LTI_CONTEXT_ID])

    return None



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import machina.models.fields
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("forum", "0001_initial"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="Post",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                (
                    "created",
                    models.DateTimeField(
                        auto_now_add=True, verbose_name="Creation date"
                    ),
                ),
                (
                    "updated",
                    models.DateTimeField(auto_now=True, verbose_name="Update date"),
                ),
                (
                    "poster_ip",
                    models.GenericIPAddressField(
                        default="2002::0",
                        null=True,
                        verbose_name="Poster IP address",
                        blank=True,
                    ),
                ),
                ("subject", models.CharField(max_length=255, verbose_name="Subject")),
                (
                    "content",
                    machina.models.fields.MarkupTextField(
                        no_rendered_field=True, verbose_name="Content"
                    ),
                ),
                (
                    "username",
                    models.CharField(
                        max_length=155, null=True, verbose_name="Username", blank=True
                    ),
                ),
                (
                    "approved",
                    models.BooleanField(default=True, verbose_name="Approved"),
                ),
                (
                    "update_reason",
                    models.CharField(
                        max_length=255,
                        null=True,
                        verbose_name="Update reason",
                        blank=True,
                    ),
                ),
                (
                    "updates_count",
                    models.PositiveIntegerField(
                        default=0,
                        verbose_name="Updates count",
                        editable=False,
                        blank=True,
                    ),
                ),
                (
                    "_content_rendered",
                    models.TextField(null=True, editable=False, blank=True),
                ),
                (
                    "poster",
                    models.ForeignKey(
                        related_name="posts",
                        verbose_name="Poster",
                        blank=True,
                        to=settings.AUTH_USER_MODEL,
                        null=True,
                        on_delete=models.CASCADE,
                    ),
                ),
            ],
            options={
                "ordering": ["created"],
                "abstract": False,
                "get_latest_by": "created",
                "verbose_name": "Post",
                "verbose_name_plural": "Posts",
            },
        ),
        migrations.CreateModel(
            name="Topic",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                (
                    "created",
                    models.DateTimeField(
                        auto_now_add=True, verbose_name="Creation date"
                    ),
                ),
                (
                    "updated",
                    models.DateTimeField(auto_now=True, verbose_name="Update date"),
                ),
                ("subject", models.CharField(max_length=255, verbose_name="Subject")),
                ("slug", models.SlugField(max_length=255, verbose_name="Slug")),
                (
                    "type",
                    models.PositiveSmallIntegerField(
                        db_index=True,
                        verbose_name="Topic type",
                        choices=[(0, "Default topic"), (1, "Sticky"), (2, "Announce")],
                    ),
                ),
                (
                    "status",
                    models.PositiveIntegerField(
                        db_index=True,
                        verbose_name="Topic status",
                        choices=[
                            (0, "Topic unlocked"),
                            (1, "Topic locked"),
                            (2, "Topic moved"),
                        ],
                    ),
                ),
                (
                    "approved",
                    models.BooleanField(default=True, verbose_name="Approved"),
                ),
                (
                    "posts_count",
                    models.PositiveIntegerField(
                        default=0,
                        verbose_name="Posts count",
                        editable=False,
                        blank=True,
                    ),
                ),
                (
                    "views_count",
                    models.PositiveIntegerField(
                        default=0,
                        verbose_name="Views count",
                        editable=False,
                        blank=True,
                    ),
                ),
                (
                    "last_post_on",
                    models.DateTimeField(
                        null=True, verbose_name="Last post added on", blank=True
                    ),
                ),
                (
                    "forum",
                    models.ForeignKey(
                        related_name="topics",
                        verbose_name="Topic forum",
                        to="forum.Forum",
                        on_delete=models.CASCADE,
                    ),
                ),
                (
                    "poster",
                    models.ForeignKey(
                        verbose_name="Poster",
                        blank=True,
                        to=settings.AUTH_USER_MODEL,
                        null=True,
                        on_delete=models.CASCADE,
                    ),
                ),
                (
                    "subscribers",
                    models.ManyToManyField(
                        related_name="subscriptions",
                        verbose_name="Subscribers",
                        to=settings.AUTH_USER_MODEL,
                        blank=True,
                    ),
                ),
            ],
            options={
                "ordering": ["-type", "-last_post_on"],
                "abstract": False,
                "get_latest_by": "last_post_on",
                "verbose_name": "Topic",
                "verbose_name_plural": "Topics",
            },
        ),
        migrations.AddField(
            model_name="post",
            name="topic",
            field=models.ForeignKey(
                related_name="posts",
                verbose_name="Topic",
                to="forum_conversation.Topic",
                on_delete=models.CASCADE,
            ),
        ),
        migrations.AddField(
            model_name="post",
            name="updated_by",
            field=models.ForeignKey(
                blank=True,
                editable=False,
                to=settings.AUTH_USER_MODEL,
                null=True,
                verbose_name="Lastly updated by",
                on_delete=models.SET_NULL,
            ),
        ),
    ]



"""This module contains django widgets related to Ashley WYSIWYG editor """
import json

from django.forms.widgets import HiddenInput


class DraftEditor(HiddenInput):
    """Draft.js editor form field widget."""

    template_name = "widgets/editor.html"

    def get_context(self, name, value, attrs):
        """Add props to context."""
        context = super().get_context(name, value, attrs)
        # add to props mentions, forum, poster information and target input used in frontend
        context["widget"]["attrs"]["target"] = (
            context.get("widget").get("attrs").get("id")
        )
        context["widget"]["props"] = json.dumps(
            context.get("widget").get("attrs"), separators=(",", ":")
        )  # nosec
        return context


__all__ = ["DraftEditor"]



"""Admin of the Ashley application."""
from django.contrib import admin

from .models import User


@admin.register(User)
class UserAdmin(admin.ModelAdmin):
    """Admin class for Ashley's User model."""



"""Views of the ashley django application."""
import logging
from typing import List

from django.conf import settings
from django.contrib import messages
from django.contrib.auth import get_user_model
from django.http import HttpResponse, HttpResponseBadRequest, HttpResponseRedirect
from django.urls import reverse
from django.utils import translation
from django.utils.translation import gettext_lazy as _
from django.views.generic import TemplateView, UpdateView
from lti_toolbox.lti import LTI
from lti_toolbox.views import BaseLTIAuthView
from machina.apps.forum_permission.shortcuts import assign_perm
from machina.apps.forum_permission.viewmixins import (
    PermissionRequiredMixin as BasePermissionRequiredMixin,
)
from machina.core.db.models import get_model
from machina.core.loading import get_class

from ashley.permissions import ManageModeratorPermission

from . import SESSION_LTI_CONTEXT_ID
from .defaults import DEFAULT_FORUM_BASE_PERMISSIONS, DEFAULT_FORUM_ROLES_PERMISSIONS

Forum = get_model("forum", "Forum")  # pylint: disable=C0103
LTIContext = get_model("ashley", "LTIContext")  # pylint: disable=C0103
PermissionRequiredMixin: BasePermissionRequiredMixin = get_class(
    "forum_permission.viewmixins", "PermissionRequiredMixin"
)
User = get_user_model()

logger = logging.getLogger(__name__)


class ForumLTIView(BaseLTIAuthView):
    """Forum view called by an LTI launch request."""

    def _do_on_login(self, lti_request: LTI) -> HttpResponse:
        """Process the request when the user is logged in via LTI"""

        # Get or create the LTIContext model associated with the current LTI launch request
        context_id = lti_request.get_param("context_id")
        if not context_id:
            return HttpResponseBadRequest("LTI parameter context_id is mandatory")
        context, _context_created = LTIContext.objects.get_or_create(
            lti_id=context_id, lti_consumer=lti_request.get_consumer()
        )

        # Synchronize the user groups related to the current LTI context
        context.sync_user_groups(self.request.user, lti_request.roles)

        # Store the current user LTI context in session
        self.request.session[SESSION_LTI_CONTEXT_ID] = context.id

        # Get or create the requested forum
        forum_uuid = self.kwargs["uuid"]
        # The requested forum must exist in this context or needs to be created
        try:
            forum = Forum.objects.get(
                lti_id=forum_uuid,
                type=Forum.FORUM_POST,
                lti_contexts__id=context.id,
            )
        except Forum.DoesNotExist:
            forum = Forum.objects.create(
                lti_id=forum_uuid,
                type=Forum.FORUM_POST,
                name=lti_request.context_title,
            )
            logger.debug("Forum %s created", forum_uuid)
            self._init_forum(forum, context)

        if not getattr(self.request.user, "public_username", ""):
            redirect_url = reverse("forum.username.change")
        else:
            redirect_url = reverse(
                "forum:forum", kwargs={"slug": forum.slug, "pk": forum.id}
            )

        response = HttpResponseRedirect(redirect_url)

        locale = lti_request.get_param("launch_presentation_locale")
        if locale:
            logger.debug("Course locale detected %s", locale)
            translation.activate(locale)
            response.set_cookie(settings.LANGUAGE_COOKIE_NAME, locale)

        return response

    def _init_forum(self, forum, context):
        """
        Initialize a forum.

        This method sets default permissions according to the LTI context that
        initiated the creation of the forum.
        """
        forum.lti_contexts.add(context)
        self._assign_permissions(
            forum,
            context.get_base_group(),
            DEFAULT_FORUM_BASE_PERMISSIONS,
        )
        # pylint: disable=no-member
        for role, perms in DEFAULT_FORUM_ROLES_PERMISSIONS.items():
            self._assign_permissions(forum, context.get_role_group(role), perms)

    @staticmethod
    def _assign_permissions(forum, group, permissions: List[str]):
        """Grant a list of permissions for a group on a specific forum."""
        for perm in permissions:
            logger.debug(
                "Grant permission %s to group %s on forum %s",
                perm,
                group.name,
                forum.lti_id,
            )
            assign_perm(perm, group, forum, True)


class ChangeUsernameView(PermissionRequiredMixin, UpdateView):
    """ Allow the current user to update his username. """

    model = User
    fields = ["public_username"]
    template_name = "ashley/change_username.html"

    def get_object(self, queryset=None):
        return self.request.user

    def form_valid(self, form):
        """If the form is valid, save the associated model."""
        # pylint: disable=attribute-defined-outside-init
        self.object = form.save()
        messages.success(
            self.request,
            _(
                "Welcome %(username)s, your username is now registered."
                % {"username": self.request.user.public_username}
            ),
        )
        return HttpResponseRedirect(reverse("forum:index"))

    # pylint: disable=no-self-use,unused-argument
    def perform_permissions_check(self, user, obj, perms):
        """ The user can update his username only if it is not yet defined """
        return user.is_authenticated and not getattr(obj, "public_username", "")


class ManageModeratorsView(ManageModeratorPermission, TemplateView):
    """View to manage revoke and promote moderators."""

    context_object_name = "users"
    template_name = "ashley/manage_moderator.html"



"""This module contains default settings for Ashley application."""

from django.conf import settings
from django.utils.functional import lazy
from draftjs_exporter.defaults import BLOCK_MAP as DEFAULT_BLOCK_MAP
from draftjs_exporter.defaults import STYLE_MAP as DEFAULT_STYLE_MAP
from draftjs_exporter.dom import DOM

from ashley.editor.decorators import emoji, image, link, mention

_FORUM_ROLE_INSTRUCTOR = "instructor"
_FORUM_ROLE_MODERATOR = "moderator"

_FORUM_BASE_PERMISSIONS = [
    "can_see_forum",
    "can_read_forum",
    "can_start_new_topics",
    "can_reply_to_topics",
    "can_edit_own_posts",
    "can_post_without_approval",
    "can_vote_in_polls",
]

_FORUM_MODERATOR_PERMISSIONS = _FORUM_BASE_PERMISSIONS + [
    "can_post_announcements",
    "can_post_stickies",
    "can_delete_own_posts",
    "can_create_polls",
    "can_lock_topics",
    "can_edit_posts",
    "can_delete_posts",
    "can_approve_posts",
    "can_reply_to_locked_topics",
]

_FORUM_ADMIN_PERMISSIONS = _FORUM_MODERATOR_PERMISSIONS + [
    "can_rename_forum",
    "can_manage_moderator",
]

DEFAULT_FORUM_BASE_PERMISSIONS = lazy(
    lambda: getattr(
        settings, "ASHLEY_DEFAULT_FORUM_BASE_PERMISSIONS", _FORUM_BASE_PERMISSIONS
    ),
    list,
)()

DEFAULT_FORUM_ROLES_PERMISSIONS = lazy(
    lambda: getattr(
        settings,
        "ASHLEY_DEFAULT_FORUM_ROLES_PERMISSIONS",
        {
            "administrator": _FORUM_ADMIN_PERMISSIONS,
            _FORUM_ROLE_INSTRUCTOR: _FORUM_ADMIN_PERMISSIONS,
            _FORUM_ROLE_MODERATOR: _FORUM_MODERATOR_PERMISSIONS,
        },
    ),
    dict,
)()

DEFAULT_DRAFTJS_EXPORTER_CONFIG = {
    "entity_decorators": {
        "LINK": link,
        "emoji": emoji,
        "mention": mention,
        "IMAGE": image,
    },
    "composite_decorators": [],
    "block_map": DEFAULT_BLOCK_MAP,
    "style_map": DEFAULT_STYLE_MAP,
    "engine": DOM.STRING,
}

DRAFTJS_EXPORTER_CONFIG = lazy(
    lambda: {
        **DEFAULT_DRAFTJS_EXPORTER_CONFIG,
        **getattr(settings, "ASHLEY_DRAFTJS_EXPORTER_CONFIG", {}),
    },
    dict,
)()


INDEXABLE_USER_DISPLAY_NAME_METHOD = getattr(
    settings, "ASHLEY_INDEXABLE_USER_DISPLAY_NAME_METHOD", "get_public_username"
)



"""Declare the models related to Ashley ."""
import logging
import pathlib
import uuid
from typing import List

from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth.models import AbstractUser as DjangoAbstractUser
from django.contrib.auth.models import Group
from django.core.exceptions import PermissionDenied
from django.core.validators import FileExtensionValidator
from django.db import models
from django.db.models import Model
from django.utils.translation import gettext_lazy as _
from lti_toolbox.models import LTIConsumer
from machina.core.db.models import get_model, model_factory
from machina.models.abstract_models import DatedModel

from .validators import validate_upload_image_file_size

logger = logging.getLogger(__name__)


class AbstractUser(DjangoAbstractUser):
    """Abstract user model for Ashley.

    The `username` and `password` fields are only used to authenticate
    django admins by django.contrib.auth.backends.ModelBackend.

    Non-admin user are authenticated via a LTI launch request, with the
    `lti_remote_identifier` and `lti_consumer` fields.

    Since the `username` field has a unique constraint, it is not used as
    public display name for users (2 users can have the same username if
    they come from distinct LTI consumers). The `public_username` field
    is used for this purpose.
    """

    lti_consumer = models.ForeignKey(LTIConsumer, on_delete=models.PROTECT, null=True)

    lti_remote_user_id = models.CharField(
        max_length=150,
        verbose_name=_("LTI remote user identifier"),
        help_text=_("Unique identifier for the user on the tool consumer"),
        blank=False,
    )

    public_username = models.CharField(
        max_length=150,
        verbose_name=_("Public username"),
        help_text=_("This username will be displayed with the user's posts"),
        blank=True,
    )

    REQUIRED_FIELDS: List[str] = []

    def get_public_username(self):
        """Getter for the public username of the user."""
        return self.public_username

    def get_public_username_with_default(self):
        """Getter for the public username of the user."""
        return self.public_username or _("Anonymous")

    class Meta:
        """Options for the ``AbstractUser`` model."""

        abstract = True
        app_label = "ashley"

        constraints = [
            models.UniqueConstraint(
                fields=["lti_consumer", "lti_remote_user_id"],
                name="ashleyuser_unique_consumer_remote_user_id",
            ),
        ]


# Default implementation of the AbstractUser, using django-machina's model factory
User = model_factory(AbstractUser)


class AbstractLTIContext(Model):
    """Abstract LTI Context model for Ashley."""

    GROUP_PREFIX = "cg"
    GROUP_DELIMITER = ":"

    lti_id = models.CharField(max_length=150, null=False, blank=False, db_index=True)

    lti_consumer = models.ForeignKey(
        LTIConsumer, on_delete=models.PROTECT, null=False, db_index=True
    )

    def get_base_group(self) -> Group:
        """
        Get the base (Django) Group for this LTI context.

        All LTI users authenticated within this LTI context should be in this group.
        """
        return self._get_or_create_group(self.base_group_name)

    @property
    def base_group_name(self) -> str:
        """
        Get the name of the base group.
        """
        return self.GROUP_DELIMITER.join([self.GROUP_PREFIX, str(self.id)])

    def get_group_role_name(self, role) -> str:
        """
        Get the name of the group for a given role
        By default it is of the form "cg:{context.id}:role:{role}"
        """
        return self.GROUP_DELIMITER.join([self.base_group_name, "role", role])

    def get_role_groups(self, roles: List[str]) -> List[Group]:
        """
        Get the Django Groups corresponding to a list of LTI Context Roles

        All LTI users authenticated within this LTI context having these roles
        should be in these groups.
        """
        return list(map(self.get_role_group, roles))

    def get_role_group(self, role: str) -> Group:
        """
        Get the Django Group corresponding to a LTI Context Role

        All LTI user authenticated within this LTI context and having this role
        should be in this group.
        """
        group_name = self.get_group_role_name(role)
        return self._get_or_create_group(group_name)

    @staticmethod
    def _get_or_create_group(group_name: str) -> Group:
        """
        Helper to get or create a Django Group by name
        """
        group, created = Group.objects.get_or_create(name=group_name)
        if created:
            logger.debug("Group %s created", group_name)
        return group

    def sync_user_groups(self, user, roles: List[str]) -> None:
        """
        Synchronize the group membership of a user for this LTI Context.
        User from LTI can sign as instructor or student. An extra group can be added
        in Ashley to promote user to moderator. This is an internal group.
        """
        # Control user is part of this lti_context
        if user.lti_consumer_id != self.lti_consumer_id:
            raise PermissionDenied(
                ("The User and LTIContext must be part of the same LTI Consumer")
            )

        current_groups = list(user.groups.filter(name__startswith=self.base_group_name))
        target_groups = self.get_role_groups(roles) + [self.get_base_group()]

        logger.debug("Current groups : %s", current_groups)
        logger.debug("Target groups : %s", target_groups)

        # Remove groups if necessary
        for group in current_groups:
            # Moderator group is an internal group, it musn't be removed
            if group not in target_groups and group.name != self.get_group_role_name(
                "moderator"
            ):
                logger.debug("Removing user %s from group %s", user, group)
                user.groups.remove(group)

        # Add group if necessary
        for group in target_groups:
            if group not in current_groups:
                logger.debug("Add user %s to group %s", user, group)
                user.groups.add(group)

    def get_user_roles(self, user) -> List[str]:
        """
        Return list of roles name without group pattern
        cg:{context.id}:role: for this user in this LTI Context
        """
        role_group_prefix = self.get_group_role_name("")
        user_role_groups = user.groups.filter(
            name__startswith=role_group_prefix
        ).values_list("name", flat=True)
        roles = [group.replace(role_group_prefix, "") for group in user_role_groups]
        return roles

    class Meta:
        """Options for the ``AbstractLTIContext`` model."""

        abstract = True
        app_label = "ashley"

        constraints = [
            models.UniqueConstraint(
                fields=["lti_consumer", "id"],
                name="lticontext_unique_consumer_context_id",
            ),
        ]


# Default implementation of the AbstractLTIContext model
LTIContext = model_factory(AbstractLTIContext)


def image_directory_path(instance, filename):
    """File will be uploaded to specific path"""
    new_filename = f"{uuid.uuid4()}{pathlib.Path(filename).suffix}"
    return f"image_uploads/{instance.forum.id}/{instance.poster.id}/{new_filename}"


class AbstractUploadImage(DatedModel):
    """Abstract model for images."""

    file = models.ImageField(
        upload_to=image_directory_path,
        max_length=100,
        blank=True,
        validators=[
            FileExtensionValidator(settings.IMAGE_TYPE_ALLOWED),
            validate_upload_image_file_size,
        ],
    )
    forum = models.ForeignKey(
        get_model("forum", "Forum"),
        on_delete=models.CASCADE,
        verbose_name=_("Forum"),
    )
    poster = models.ForeignKey(
        get_user_model(),
        on_delete=models.CASCADE,
        verbose_name=_("Poster"),
    )

    class Meta:
        """Options for the `AbstractImage` model."""

        abstract = True
        app_label = "ashley"


# Default implementation of the AbstractImage model
UploadImage = model_factory(AbstractUploadImage)



# Generated by Django 3.1.4 on 2020-12-14 17:50

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("ashley", "0003_auto_20200917_1335"),
    ]

    operations = [
        migrations.AlterField(
            model_name="user",
            name="public_username",
            field=models.CharField(
                blank=True,
                help_text="This username will be displayed with the user's posts",
                max_length=150,
                verbose_name="Public username",
            ),
        ),
    ]



"""
    Forum search indexes
    ====================
    This module defines search indexes allowing to perform searches among forum topics and posts.
"""

from haystack import indexes
from machina.core.db.models import get_model

from ashley import defaults

Post = get_model("forum_conversation", "Post")


def get_indexable_forum_member_display_name(user):
    """ Given a specific user, returns their related display name. """
    return getattr(user, defaults.INDEXABLE_USER_DISPLAY_NAME_METHOD)()


class PostIndex(indexes.SearchIndex, indexes.Indexable):
    """
    Defines the data stored in the Post indexes.

    Derives from Django Machina's PostIndex to define "forum_name" and "topic_name" as Edge Ngrams
    """

    text = indexes.EdgeNgramField(
        document=True,
        use_template=True,
        template_name="forum_search/post_text.txt",
    )

    poster = indexes.IntegerField(model_attr="poster_id", null=True)
    poster_name = indexes.EdgeNgramField(null=True)

    forum = indexes.IntegerField(model_attr="topic__forum_id")
    forum_slug = indexes.CharField()
    forum_name = indexes.EdgeNgramField()

    topic = indexes.IntegerField(model_attr="topic_id")
    topic_slug = indexes.CharField()
    topic_subject = indexes.EdgeNgramField()

    created = indexes.DateTimeField(model_attr="created")
    updated = indexes.DateTimeField(model_attr="updated")

    def get_model(self):
        return Post

    @staticmethod
    def prepare_poster_name(obj):
        """ Returns the poster's name """
        return (
            get_indexable_forum_member_display_name(obj.poster)
            if obj.poster
            else obj.username
        )

    @staticmethod
    def prepare_forum_slug(obj):
        """ Returns the forum's slug """
        return obj.topic.forum.slug

    @staticmethod
    def prepare_forum_name(obj):
        """ Returns the forum's name """
        return obj.topic.forum.name

    @staticmethod
    def prepare_topic_slug(obj):
        """ Returns the topic's slug """
        return obj.topic.slug

    @staticmethod
    def prepare_topic_subject(obj):
        """ Returns the topic's subject """
        return obj.topic.subject

    def index_queryset(self, using=None):
        return Post.objects.all().exclude(approved=False)

    def read_queryset(self, using=None):
        return (
            Post.objects.all().exclude(approved=False).select_related("topic", "poster")
        )



# Generated by Django 3.1.2 on 2020-12-20 22:45

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("forum_conversation", "0012_auto_20200423_1049"),
    ]

    operations = [
        migrations.AlterField(
            model_name="topic",
            name="created",
            field=models.DateTimeField(
                auto_now_add=True, db_index=True, verbose_name="Creation date"
            ),
        ),
        migrations.AlterField(
            model_name="topic",
            name="last_post_on",
            field=models.DateTimeField(
                blank=True, db_index=True, null=True, verbose_name="Last post added on"
            ),
        ),
        migrations.AlterField(
            model_name="topic",
            name="updated",
            field=models.DateTimeField(
                auto_now=True, db_index=True, verbose_name="Update date"
            ),
        ),
        migrations.AddIndex(
            model_name="topic",
            index=models.Index(
                fields=["type", "last_post_on"], name="forum_conve_type_cc96d0_idx"
            ),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.6 on 2016-06-07 03:00
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("forum_permission", "0001_initial"),
    ]

    operations = [
        migrations.AlterField(
            model_name="forumpermission",
            name="codename",
            field=models.CharField(
                db_index=True,
                max_length=150,
                unique=True,
                verbose_name="Permission codename",
            ),
        ),
        migrations.AlterField(
            model_name="forumpermission",
            name="is_global",
            field=models.BooleanField(
                db_index=True,
                default=False,
                help_text="This permission can be granted globally to all the forums",
                verbose_name="Global permission",
            ),
        ),
        migrations.AlterField(
            model_name="forumpermission",
            name="is_local",
            field=models.BooleanField(
                db_index=True,
                default=True,
                help_text="This permission can be granted individually for each forum",
                verbose_name="Local permission",
            ),
        ),
        migrations.AlterField(
            model_name="groupforumpermission",
            name="has_perm",
            field=models.BooleanField(
                db_index=True, default=True, verbose_name="Has perm"
            ),
        ),
        migrations.AlterField(
            model_name="userforumpermission",
            name="anonymous_user",
            field=models.BooleanField(
                db_index=True, default=False, verbose_name="Target anonymous user"
            ),
        ),
        migrations.AlterField(
            model_name="userforumpermission",
            name="has_perm",
            field=models.BooleanField(
                db_index=True, default=True, verbose_name="Has perm"
            ),
        ),
    ]



"""This module contains custom decorators for draftjs_exporter."""
import logging

from draftjs_exporter.dom import DOM

logger = logging.getLogger(__name__)


def link(props):
    """
    Decorator for the `LINK` entity in Draft.js ContentState.

    `draftjs_exporter` does not provide a decorator for this kind of entity by
     default, so we have to define it to support links in forum posts.
    """
    title = props.get("title")
    href = props.get("url", "#")
    anchor_properties = {
        "href": href,
        "target": "_blank",
        "rel": "nofollow noopener noreferrer",
    }

    if title is not None:
        anchor_properties["title"] = title

    return DOM.create_element("a", anchor_properties, props["children"])


def emoji(props):
    """
    Decorator for the `emoji` entity in Draft.js ContentState.

    This entity is added by the `draft-js-emoji-plugin` plugin.
    (https://www.draft-js-plugins.com/plugin/emoji)
    """
    return DOM.create_element("span", {"class": "emoji"}, props["children"])


def mention(props):
    """
    Decorator for the `mention` entity in Draft.js ContentState.

    This entity is added by the `draft-js-mention-plugin` plugin.
    (https://www.draft-js-plugins.com/plugin/mention)
    """

    user_id = props.get("mention").get("user")
    name = props.get("mention").get("name")

    if name and user_id:
        return DOM.create_element(
            "a",
            {
                "class": "mention",
                "href": f"/forum/member/profile/{user_id}/",
            },
            DOM.create_element("span", {"class": "mention"}, f"@{name}"),
        )

    return None


def image(props):
    """
    Decorator for the `IMAGE` entity in Draft.js ContentState. Image can be decorated with a width,
    height, and alignment property. By default draft-js insert data at 40% width of the editor,
    we want to keep the same ratio so that the preview is alike.
    """
    css_class = ""
    alignment = props.get("alignment", None)
    if alignment in ("left", "right"):
        css_class = f"float-{alignment}"
    if alignment == "center":
        css_class = "image-center"

    # make sure data are number and positive, default is 40% of editor size
    width = f'{abs(int(props.get("width", "40")))}%'

    height = "auto"
    if props.get("height", None) is not None:
        height = f'{abs(int(props.get("height")))}%'

    if props.get("src", None) is not None:
        return DOM.create_element(
            "img",
            {
                "class": css_class,
                "src": props.get("src"),
                "width": width,
                "height": height,
            },
        )

    return None
